# encoding: utf-8
#
# Author::    Paweł Wilk (mailto:pw@gnu.org)
# Copyright:: (c) 2011,2012,2013 by Paweł Wilk
# License::   This program is licensed under the terms of {file:docs/LGPL GNU Lesser General Public License} or {file:docs/COPYING Ruby License}.
# 
# This file contains version information.

module I18n
  module Inflector

    # @private
    DEVELOPER   = 'Paweł Wilk'
    # @private
    EMAIL       = 'pw@gnu.org'
    # @private
    VERSION     = '2.6.7'
    # @private
    NAME        = 'i18n-inflector'
    # @private
    SUMMARY     = 'Inflection module for I18n'
    # @private
    URL         = 'https://rubygems.org/gems/i18n-inflector/'
    # @private
    DESCRIPTION = 'Enhances simple I18n backend in a way that it inflects translation data using pattern interpolation.'

  end
end
